# -*- coding: utf-8 -*-

"""INTERFACE ROUTINES

This module contains submodules for error handling, logging and IO interaction.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.0

:Date: 06/04/2017

"""
