# -*- coding: utf-8 -*-

"""BASE ROUTINES

This module contains submodules for basic operations such as type
transformations and adjustments to the default output of Numpy functions.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.0

:Date: 06/04/2017

"""
