# -*- coding: utf-8 -*-

"""MATHEMATICS ROUTINES

This module contains submodules for mathematical applications.

:Author: Samuel Farrens <samuel.farrens@gmail.com>

:Version: 1.0

:Date: 06/04/2017

"""
